//
//  Copyright (C) 2002 Constantin Kaplinsky, Inc.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

//
// HTTPConnectSocket.java together with HTTPConnectSocketFactory.java
// implement an alternate way to connect to VNC servers via one or two
// HTTP proxies supporting the HTTP CONNECT method.
//
// --------------------------------------------------------------------
// Modified on 16-Apr-2003 by Warren Knight
// To support Basic Authentication from the HTTP Proxy Server
// --------------------------------------------------------------------
package nxm.vnc.libvnc.tmde;
import nxm.vnc.libvnc.tightvnc.*;


import java.net.*;
import java.io.*;

class HTTPConnectSocketAuth extends Socket {
  protected boolean authenticationRequired = false;
  protected String authscheme;



  public String getAuthScheme() {
    return authscheme;
  }

  public boolean isAuthenticationRequired() {
    return authenticationRequired;
  }

  public HTTPConnectSocketAuth(String host, int port,
			   String proxyHost, int proxyPort)
    throws IOException {
    // Connect to the specified HTTP proxy
    super(proxyHost, proxyPort);
    initialize(host,port,proxyHost,proxyPort,null,null);
  }



  public HTTPConnectSocketAuth(String host, int port,
			   String proxyHost, int proxyPort,
                           String inauthscheme, String credentials)
    throws IOException {
    // Connect to the specified HTTP proxy
    super(proxyHost, proxyPort);
    initialize(host,port,proxyHost,proxyPort,inauthscheme,credentials);
  }


  public void initialize(String host, int port,
	            String proxyHost, int proxyPort,
                    String inauthscheme, String credentials)
    throws IOException {
    String connectRequest;
    if (inauthscheme == null) {
     // Send the CONNECT request
     connectRequest="CONNECT " + host + ":" + port + " HTTP/1.0\r\n\r\n";
    } else {
     // Send the CONNECT request with authorization
     connectRequest="CONNECT " + host + ":" + port + " HTTP/1.0\r\n" +
                   "Proxy-Authenticate: " + inauthscheme + "\r\n" +
                   "Proxy-Authorization: Basic " + credentials + "\r\n\r\n";
    }
    getOutputStream().write(connectRequest.getBytes());

    // Read the first line of the response
//  DataInputStream is = new DataInputStream(getInputStream()); // ORIG prior to 2.9.2
    BufferedReader is = new BufferedReader(new InputStreamReader(getInputStream())); // since 2.9.2 to fix Deprecated warning

    String str = is.readLine();

    // Check to see if Authentication is required
    // Note Chained Proxies could cause this several times
    if (str.startsWith("HTTP/1.0 407 ") || str.startsWith("HTTP/1.1 407 ")) {
      authenticationRequired = true;
      System.out.println(str);
      do {
        str = is.readLine();
        if (str.startsWith("Proxy-Authenticate:")) {
         authscheme = str.substring(20);
         System.out.println("Determine AuthScheme to be: " +authscheme);
        }
        System.out.println(str);
      } while (str.length() != 0);
      close(); // This Socket is of no use now close it
               // so the factory knows to try again
    } else {

     // Check the HTTP error code -- it should be "200" on success
     if (!str.startsWith("HTTP/1.0 200 ") && !str.startsWith("HTTP/1.1 200 ")) {
       System.out.println(str);
       do {
         str = is.readLine();
         System.out.println(str);
       } while (str.length() != 0);
       throw new IOException("Proxy Connect Failed");
     }

     // Success -- skip remaining HTTP headers
     do {
       str = is.readLine();
       System.out.println(str);
     } while (str.length() != 0);
   }
 }
}

