package nxm.sys.inc;

import java.lang.annotation.Documented;
import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;

/** Marker annotation for Java code that is for <b>INTERNAL USE ONLY</b>. Any
    code so marked may be changed or removed at any time and may have
    dangerous un-checked and/or un-documented behavior if used incorrectly.
    @author  Neon Ngo
    @since   NeXtMidas 3.3.0
 */
@Documented
@Inherited
@Retention(RetentionPolicy.RUNTIME)
//@Target is specifically omitted as omitting it permits all targets
@InternalUseOnly
public @interface InternalUseOnly {
  /** Optional description or notes on when, why, etc. marked item is for internal use only.
      @return description or notes on why this tag is for internal use only
      @since 3.5.1
   */
  String value() default "";
}
